from .asset_list import AssetList
from .blend_tile_data import BlendTileData
from .camera_animation_list import CameraAnimationList
from .environment_data import EnvironmentData
from .global_lighting import GlobalLighting
from .global_version import GlobalVersion
from .height_map import HeightMapData
from .library_map_lists import LibraryMapLists
from .mp_positions import MPPositionsList
from .named_cameras import NamedCameras
from .object_list import ObjectsList
from .player_scripts import PlayerScriptsList
from .polygon_triggers import PolygonTriggers
from .post_effects_chunk import PostEffectsChunk
from .river_areas import RiverAreas
from .sides_list import BuildLists, SidesList
from .skipped_asset import SkippedAsset
from .standing_water_area import StandingWaterAreas
from .standing_waves_area import StandingWaveAreas
from .teams import Teams
from .trigger_areas import TriggerAreas
from .water_settings import WaterSettings
from .waypoint_list import WaypointsList
from .world_info import WorldInfo

__all__ = [
    "WaterSettings",
    "HeightMapData",
    "WorldInfo",
    "BlendTileData",
    "GlobalVersion",
    "SkippedAsset",
    "WaypointsList",
    "Teams",
    "MPPositionsList",
    "ObjectsList",
    "PlayerScriptsList",
    "PolygonTriggers",
    "SidesList",
    "TriggerAreas",
    "BuildLists",
    "AssetList",
    "StandingWaterAreas",
    "StandingWaveAreas",
    "RiverAreas",
    "GlobalLighting",
    "EnvironmentData",
    "PostEffectsChunk",
    "NamedCameras",
    "CameraAnimationList",
    "LibraryMapLists",
]



"""Test NamedCameras asset parsing."""

from sagemap.assets import NamedCameras
from .conftest import create_context, load_asset_bytes


def test_named_cameras():
    """Test NamedCameras asset parsing."""
    asset_bytes = load_asset_bytes("NamedCameras")
    
    context = create_context(asset_bytes)
    result = NamedCameras.parse(context)
    assert result is not None



from dataclasses import dataclass

from sagemap.context import ParsingContext, Property


@dataclass
class WorldInfo:
    asset_name = "WorldInfo"

    version: int
    properties: dict[str, "Property"]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            properties = context.properties_to_dict(context.parse_properties())

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, properties)



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class WaterSettings:
    asset_name = "GlobalWaterSettings"

    version: int
    reflection_on: bool
    reflection_plane_z: float

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            reflection_on = context.stream.readBool()
            reflection_plane_z = context.stream.readSingle()

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, reflection_on, reflection_plane_z)



import enum
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


class TimeOfTheDay(enum.Enum):
    Morning = 1
    Afternoon = 2
    Evening = 3
    Night = 4

    def map_to_str(self) -> str:
        times = {
            TimeOfTheDay.Morning: "MORNING",
            TimeOfTheDay.Afternoon: "AFTERNOON",
            TimeOfTheDay.Evening: "EVENING",
            TimeOfTheDay.Night: "NIGHT",
        }

        return times[self]


@dataclass
class MapColorArgb:
    a: int
    r: int
    g: int
    b: int

    @classmethod
    def parse(cls, context: "ParsingContext"):
        value = context.stream.readUInt32()
        a = (value >> 24) & 0xFF
        r = (value >> 16) & 0xFF
        g = (value >> 8) & 0xFF
        b = value & 0xFF

        return cls(a=a, r=r, g=g, b=b)


@dataclass
class GlobalLight:
    ambient: tuple[float, float, float]
    color: tuple[float, float, float]
    direction: tuple[float, float, float]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        ambient = context.stream.readVector3()
        color = context.stream.readVector3()
        direction = context.stream.readVector3()

        return cls(ambient, color, direction)


@dataclass
class GlobalLightingConfiguration:
    terrain_sun: GlobalLight
    object_sun: GlobalLight | None
    infantry_sun: GlobalLight | None
    terrain_accent1: GlobalLight
    object_accent1: GlobalLight | None
    infantry_accent1: GlobalLight | None
    terrain_accent2: GlobalLight
    object_accent2: GlobalLight | None
    infantry_accent2: GlobalLight | None

    @classmethod
    def parse(cls, context: "ParsingContext", version: int):
        terrain_sun = GlobalLight.parse(context)

        object_sun = None
        infantry_sun = None
        if version < 10:
            object_sun = GlobalLight.parse(context)

            if version >= 7:
                infantry_sun = GlobalLight.parse(context)

        terrain_accent1 = GlobalLight.parse(context)

        object_accent1 = None
        infantry_accent1 = None
        if version < 10:
            object_accent1 = GlobalLight.parse(context)

            if version >= 7:
                infantry_accent1 = GlobalLight.parse(context)

        terrain_accent2 = GlobalLight.parse(context)

        object_accent2 = None
        infantry_accent2 = None
        if version < 10:
            object_accent2 = GlobalLight.parse(context)

            if version >= 7:
                infantry_accent2 = GlobalLight.parse(context)

        return cls(
            terrain_sun=terrain_sun,
            object_sun=object_sun,
            infantry_sun=infantry_sun,
            terrain_accent1=terrain_accent1,
            object_accent1=object_accent1,
            infantry_accent1=infantry_accent1,
            terrain_accent2=terrain_accent2,
            object_accent2=object_accent2,
            infantry_accent2=infantry_accent2,
        )


@dataclass
class GlobalLighting:
    asset_name = "GlobalLighting"

    time_of_the_day: TimeOfTheDay
    lighting_configurations: dict[TimeOfTheDay, GlobalLightingConfiguration]
    shadow_color: MapColorArgb
    unknown: bytes | None
    unknown2: tuple[float, float, float] | None
    unknown3: MapColorArgb | None
    no_cloud_factor: tuple[float, float, float] | None

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            time = TimeOfTheDay(context.stream.readUInt32())
            lighting_configurations = {}

            for member in TimeOfTheDay:
                lighting_configurations[member] = GlobalLightingConfiguration.parse(context, version)

            shadow_color = MapColorArgb.parse(context)

            unknown = None
            if version >= 7 and version < 11:
                unknown = context.stream.readBytes(4 if version >= 9 else 44)

            unknown2 = None
            unknown3 = None
            if version >= 12:
                unknown2 = context.stream.readVector3()
                unknown3 = MapColorArgb.parse(context)

            no_cloud_factor = None
            if version >= 8:
                no_cloud_factor = context.stream.readVector3()

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(
            time_of_the_day=time,
            lighting_configurations=lighting_configurations,
            shadow_color=shadow_color,
            unknown=unknown,
            unknown2=unknown2,
            unknown3=unknown3,
            no_cloud_factor=no_cloud_factor,
        )



"""Test CameraAnimationList asset parsing."""

from sagemap.assets import CameraAnimationList
from .conftest import create_context, load_asset_bytes


def test_camera_animation_list():
    """Test CameraAnimationList asset parsing."""
    asset_bytes = load_asset_bytes("CameraAnimationList")
    
    context = create_context(asset_bytes)
    result = CameraAnimationList.parse(context)
    assert result is not None



from dataclasses import dataclass

from sagemap.context import ParsingContext, Property


@dataclass
class Team:
    properties: dict[str, "Property"]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        properties = context.properties_to_dict(context.parse_properties())
        context.logger.debug(f"Parsed Team with properties: {properties}")
        return cls(properties=properties)


@dataclass
class Teams:
    asset_name = "Teams"

    version: int
    teams: list[Team]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            teams = []
            team_count = context.stream.readUInt32()
            for _ in range(team_count):
                teams.append(Team.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version=version, teams=teams)



"""Test StandingWaveAreas asset parsing."""

from sagemap.assets import StandingWaveAreas
from .conftest import create_context, load_asset_bytes


def test_standing_wave_areas():
    """Test StandingWaveAreas asset parsing."""
    asset_bytes = load_asset_bytes("StandingWaveAreas")
    
    context = create_context(asset_bytes)
    result = StandingWaveAreas.parse(context)
    assert result is not None



"""Test WorldInfo asset parsing."""

from sagemap.assets import WorldInfo
from .conftest import create_context, load_asset_bytes


def test_world_info():
    """Test WorldInfo asset parsing."""
    asset_bytes = load_asset_bytes("WorldInfo")
    
    context = create_context(asset_bytes)
    result = WorldInfo.parse(context)
    assert result is not None



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class NamedCamera:
    look_at_point: tuple[float, float, float]
    name: str
    pitch: float
    roll: float
    yaw: float
    zoom: float
    fov: float
    unknown: float

    @classmethod
    def parse(cls, context: "ParsingContext"):
        look_at_point = context.stream.readVector3()
        name = context.stream.readUInt16PrefixedAsciiString()
        pitch = context.stream.readFloat()
        roll = context.stream.readFloat()
        yaw = context.stream.readFloat()
        zoom = context.stream.readFloat()
        fov = context.stream.readFloat()
        unknown = context.stream.readFloat()

        context.logger.debug(
            f"NamedCamera: {name}, LookAt: {look_at_point}, Pitch: {pitch}, Roll: {roll}, Yaw: {yaw}, Zoom: {zoom}, FOV: {fov}"
        )
        return cls(
            look_at_point=look_at_point,
            name=name,
            pitch=pitch,
            roll=roll,
            yaw=yaw,
            zoom=zoom,
            fov=fov,
            unknown=unknown,
        )


@dataclass
class NamedCameras:
    asset_name = "NamedCameras"

    version: int
    cameras: list[NamedCamera]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            cameras = []
            camera_count = context.stream.readUInt32()
            for _ in range(camera_count):
                cameras.append(NamedCamera.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version=version, cameras=cameras)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from .teams import Team

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class BuildListInfo:
    build_name: str
    template_name: str
    location: tuple[float, float, float]
    angle: float
    is_initially_built: bool
    num_rebuilds: int
    script: str
    health: int
    whiner: bool
    unsellable: bool
    repairable: bool
    unknown: bool | None = None

    @classmethod
    def parse(cls, context: "ParsingContext", version: int, has_asset_list: bool):
        build_name = context.stream.readUInt16PrefixedAsciiString()
        template_name = context.stream.readUInt16PrefixedAsciiString()
        location = context.stream.readVector3()
        angle = context.stream.readFloat()
        is_initially_built = context.stream.readBool()

        unknown = None
        if version >= 6 and has_asset_list:
            unknown = context.stream.readBool()

        num_rebuilds = context.stream.readUInt32()
        script = context.stream.readUInt16PrefixedAsciiString()
        health = context.stream.readInt32()
        whiner = context.stream.readBool()
        unsellable = context.stream.readBool()
        repairable = context.stream.readBool()

        context.logger.debug(f"BuildListItem: {build_name}, Template: {template_name}, Health: {health}")

        return cls(
            build_name=build_name,
            template_name=template_name,
            location=location,
            angle=angle,
            is_initially_built=is_initially_built,
            num_rebuilds=num_rebuilds,
            script=script,
            health=health,
            whiner=whiner,
            unsellable=unsellable,
            repairable=repairable,
            unknown=unknown,
        )


@dataclass
class BuildList:
    asset_name = "BuildList"

    faction_name: str
    faction_name_property: tuple[int, str] | None
    build_list: list[BuildListInfo]

    @classmethod
    def parse(cls, context: "ParsingContext", version: int, has_asset_list: bool):
        faction_name = None
        faction_name_property = None
        if has_asset_list:
            faction_name = context.stream.readUInt16PrefixedAsciiString()
        else:
            faction_name_property = context.parse_asset_property_key()

        build_list = []
        build_list_count = context.stream.readUInt32()
        for _ in range(build_list_count):
            build_list.append(BuildListInfo.parse(context, version, has_asset_list))

        context.logger.debug(
            f"Parsed BuildList with {len(build_list)} items for faction '{faction_name or faction_name_property}'"
        )
        return cls(
            faction_name=faction_name,
            faction_name_property=faction_name_property,
            build_list=build_list,
        )


@dataclass
class BuildLists:
    asset_name = "BuildLists"

    version: int
    build_lists: list[BuildList]

    @classmethod
    def parse(cls, context: "ParsingContext", has_asset_list: bool):
        with context.read_asset() as (version, _):
            build_list_count = context.stream.readUInt32()
            build_lists = []
            for _ in range(build_list_count):
                item = BuildList.parse(context, version, has_asset_list)
                build_lists.append(item)

        context.logger.debug("Finished parsing BuildLists")
        return cls(version, build_lists)


@dataclass
class Player:
    properties: dict
    build_list_items: dict[str, BuildListInfo]

    @classmethod
    def parse(cls, context: "ParsingContext", version: int, has_asset_list: bool):
        properties = context.properties_to_dict(context.parse_properties())

        build_list_count = context.stream.readUInt32()
        build_lists = {}
        for _ in range(build_list_count):
            item = BuildListInfo.parse(context, version, has_asset_list)
            build_lists[item.build_name] = item

        context.logger.debug(f"Parsed Side with {len(build_lists)} build list items")

        return cls(properties=properties, build_list_items=build_lists)


@dataclass
class SidesList:
    asset_name = "SidesList"

    version: int
    unknown1: bool
    players: list[Player]
    teams: list[Team] = field(default_factory=list)

    @classmethod
    def parse(cls, context: "ParsingContext", has_asset_list: bool):
        with context.read_asset() as (version, datasize):
            end_pos = context.stream.tell() + datasize

            unknown1 = False
            if version >= 6:
                unknown1 = context.stream.readBool()

            player_count = context.stream.readUInt32()
            players = []
            for _ in range(player_count):
                players.append(Player.parse(context, version, has_asset_list))

            if version >= 5:
                teams = []
            else:
                teams = []
                if version >= 2:
                    team_count = context.stream.readUInt32()
                    for _ in range(team_count):
                        teams.append(Team.parse(context))

                while context.stream.tell() < end_pos:
                    asset_name = context.parse_asset_name()
                    if asset_name == "Team":
                        raise ValueError("Unexpected Team asset in SidesList")
                    else:
                        raise ValueError(f"Unexpected asset in {cls.asset_name}: {asset_name}")

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version=version, unknown1=unknown1, players=players, teams=teams)



"""Test WaterSettings asset parsing."""
import pytest 

from sagemap.assets import WaterSettings
from .conftest import create_context, load_asset_bytes


@pytest.mark.skip(reason="No test data available for AssetList yet")
def test_water_settings():
    """Test WaterSettings asset parsing."""
    asset_bytes = load_asset_bytes("WaterSettings")
    
    context = create_context(asset_bytes)
    result = WaterSettings.parse(context)
    assert result is not None



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext, Property


@dataclass
class Object:
    asset_name = "Object"

    version: int
    position: tuple[float, float, float]
    angle: float
    road_type: int
    type_name: str
    properties: dict[str, "Property"]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            position = context.stream.readVector3()
            angle = context.stream.readFloat()
            road_type = context.stream.readUInt32()
            type_name = context.stream.readUInt16PrefixedAsciiString()
            properties = context.properties_to_dict(context.parse_properties())

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, position, angle, road_type, type_name, properties)


@dataclass
class ObjectsList:
    asset_name = "ObjectsList"

    version: int
    object_list: list[Object]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, datasize):
            object_list = []
            end_pos = context.stream.tell() + datasize
            while context.stream.tell() < end_pos:
                asset_name = context.parse_asset_name()
                if asset_name != Object.asset_name:
                    raise ValueError(f"Expected {Object.asset_name} asset, got {asset_name}")

                object_list.append(Object.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, object_list)



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class PolygonTrigger:
    asset_name = "PolygonTrigger"

    name: str
    layer_name: str | None
    trigger_id: int
    is_water: bool
    is_river: bool
    river_start: bool | None
    river_texture: str | None
    noise_texture: str | None
    alpha_edge_texture: str | None
    sparkle_texture: str | None
    bump_map_texture: str | None
    sky_texture: str | None
    use_additive_blending: bool
    river_color: tuple[int, int, int] | None
    unknown: int | None
    uv_scroll_speed: tuple[float, float] | None
    river_alpha: float | None
    points: list[tuple[int, int, int]]

    @classmethod
    def parse(cls, context: "ParsingContext", version: int):
        name = context.stream.readUInt16PrefixedAsciiString()

        layer_name = None
        if version >= 4:
            layer_name = context.stream.readUInt16PrefixedAsciiString()

        trigger_id = context.stream.readUInt32()

        is_water = False
        if version >= 2:
            is_water = context.stream.readBool()

        is_river = False
        river_start = None
        if version >= 3:
            is_river = context.stream.readBool()
            river_start = context.stream.readBoolUInt32()

        river_texture = None
        noise_texture = None
        alpha_edge_texture = None
        sparkle_texture = None
        bump_map_texture = None
        sky_texture = None
        use_additive_blending = False
        river_color = None
        unknown = None
        uv_scroll_speed = None
        river_alpha = None
        if version >= 5:
            river_texture = context.stream.readUInt16PrefixedAsciiString()
            noise_texture = context.stream.readUInt16PrefixedAsciiString()
            alpha_edge_texture = context.stream.readUInt16PrefixedAsciiString()
            sparkle_texture = context.stream.readUInt16PrefixedAsciiString()
            bump_map_texture = context.stream.readUInt16PrefixedAsciiString()
            sky_texture = context.stream.readUInt16PrefixedAsciiString()
            use_additive_blending = context.stream.readBool()
            river_color = (
                context.stream.readUChar(),
                context.stream.readUChar(),
                context.stream.readUChar(),
            )
            unknown = context.stream.readUChar()
            uv_scroll_speed = context.stream.readVector2()
            river_alpha = context.stream.readFloat()

        point_count = context.stream.readUInt32()
        points = []
        for _ in range(point_count):
            points.append(
                (
                    context.stream.readInt32(),
                    context.stream.readInt32(),
                    context.stream.readInt32(),
                )
            )

        return cls(
            name,
            layer_name,
            trigger_id,
            is_water,
            is_river,
            river_start,
            river_texture,
            noise_texture,
            alpha_edge_texture,
            sparkle_texture,
            bump_map_texture,
            sky_texture,
            use_additive_blending,
            river_color,
            unknown,
            uv_scroll_speed,
            river_alpha,
            points,
        )


@dataclass
class PolygonTriggers:
    asset_name = "PolygonTriggers"

    version: int
    polygon_triggers: list[PolygonTrigger]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            polygon_triggers = []

            trigger_count = context.stream.readUInt32()
            max_trigger_id = 0
            for _ in range(trigger_count):
                trigger = PolygonTrigger.parse(context, version)
                polygon_triggers.append(trigger)

                if trigger.trigger_id > max_trigger_id:
                    max_trigger_id = trigger.trigger_id

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, polygon_triggers)



"""Test GlobalVersion asset parsing."""

import pytest
from sagemap.assets import GlobalVersion
from .conftest import create_context, load_asset_bytes

    
@pytest.mark.skip(reason="No test data available for AssetList yet")
def test_global_version():
    """Test GlobalVersion asset parsing."""
    asset_bytes = load_asset_bytes("GlobalVersion")
    
    context = create_context(asset_bytes)
    result = GlobalVersion.parse(context)
    assert result is not None
    assert hasattr(result, 'version')



"""Test PostEffectsChunk asset parsing."""

from sagemap.assets import PostEffectsChunk
from .conftest import create_context, load_asset_bytes


def test_post_effects_chunk():
    """Test PostEffectsChunk asset parsing."""
    asset_bytes = load_asset_bytes("PostEffectsChunk")
    
    context = create_context(asset_bytes)
    result = PostEffectsChunk.parse(context)
    assert result is not None



"""Test PolygonTriggers asset parsing."""

import pytest
from sagemap.assets import PolygonTriggers
from .conftest import create_context, load_asset_bytes


@pytest.mark.skip(reason="No test data available for AssetList yet")
def test_polygon_triggers():
    """Test PolygonTriggers asset parsing."""
    asset_bytes = load_asset_bytes("PolygonTriggers")
    
    context = create_context(asset_bytes)
    result = PolygonTriggers.parse(context)
    assert result is not None



"""Test full map parsing for all maps in tests/data/maps/."""

import pytest
from pathlib import Path
from sagemap import parse_map_from_path


def get_test_maps():
    """Get all map files from tests/data/maps/ directory."""
    maps_dir = Path(__file__).parent / "data" / "maps"
    if not maps_dir.exists():
        return []
    return list(maps_dir.glob("*.map"))


@pytest.mark.parametrize("map_path", get_test_maps(), ids=lambda p: p.name)
def test_parse_map(map_path):
    """Test full parsing of individual map files.
    
    This test ensures that each map in the test data directory can be:
    1. Successfully parsed without errors
    2. Produces a valid Map object
    3. Can be converted to a dictionary (for serialization)
    """
    # Parse the map
    map_obj = parse_map_from_path(str(map_path))
    
    # Verify we got a valid map object
    assert map_obj is not None, f"Failed to parse map: {map_path.name}"
    
    # Verify the map can be converted to dict (tests serialization)
    map_dict = map_obj.to_dict()
    assert map_dict is not None, f"Failed to convert map to dict: {map_path.name}"
    assert isinstance(map_dict, dict), f"to_dict() did not return a dict: {map_path.name}"


def test_maps_directory_exists():
    """Verify that the maps test data directory exists."""
    maps_dir = Path(__file__).parent / "data" / "maps"
    assert maps_dir.exists(), "tests/data/maps/ directory does not exist"


def test_has_test_maps():
    """Verify that there are map files to test."""
    maps = get_test_maps()
    assert len(maps) > 0, "No .map files found in tests/data/maps/ directory"



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext, WritingContext


@dataclass
class GlobalVersion:
    asset_name = "GlobalVersion"

    version: int

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            pass

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version=version)

    def write(self, context: "WritingContext"):
        with context.write_asset(self.asset_name, self.version):
            pass



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class TriggerArea:
    asset_name = "TriggerArea"

    name: str
    layer_name: str
    area_id: int
    points: list[tuple[float, float]]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        name = context.stream.readUInt16PrefixedAsciiString()
        layer_name = context.stream.readUInt16PrefixedAsciiString()
        area_id = context.stream.readUInt32()

        point_count = context.stream.readUInt32()
        points = []
        for _ in range(point_count):
            points.append(context.stream.readVector2())

        unknown2 = context.stream.readUInt32()
        if unknown2 != 0:
            raise ValueError("Invalid data in 'unknown2'")

        return cls(name, layer_name, area_id, points)


@dataclass
class TriggerAreas:
    asset_name = "TriggerAreas"

    version: int
    trigger_areas: list[TriggerArea]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            trigger_areas = []
            area_count = context.stream.readUInt32()
            for _ in range(area_count):
                trigger_areas.append(TriggerArea.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, trigger_areas)



"""Test HeightMapData asset parsing."""

from sagemap.assets import HeightMapData
from .conftest import create_context, load_asset_bytes


def test_height_map_data():
    """Test HeightMapData asset parsing."""
    asset_bytes = load_asset_bytes("HeightMapData")
    
    context = create_context(asset_bytes)
    result = HeightMapData.parse(context)
    assert result is not None



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class FreeCameraAnimationCameraFrame:
    frame_index: int
    interpolation_type: str
    position: tuple[float, float, float]
    rotation: tuple[float, float, float, float]
    fov: float

    @classmethod
    def parse(cls, context: "ParsingContext"):
        frame_index = context.stream.readUInt32()
        interpolation_type = context.stream.readFourCc()[::-1]  # Big endian, so reverse
        if interpolation_type not in ["catm", "line"]:
            raise ValueError(f"Invalid interpolation type: {interpolation_type}")

        position = context.stream.readVector3()
        rotation = context.stream.readVector4()
        fov = context.stream.readFloat()

        return cls(
            frame_index=frame_index,
            interpolation_type=interpolation_type,
            position=position,
            rotation=rotation,
            fov=fov,
        )


@dataclass
class FreeCameraAnimationFrameData:
    frames: list[FreeCameraAnimationCameraFrame]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        camera_frames_count = context.stream.readUInt32()
        camera_frames = []
        for _ in range(camera_frames_count):
            camera_frames.append(FreeCameraAnimationCameraFrame.parse(context))

        return cls(frames=camera_frames)


@dataclass
class LookAtCameraAnimationLookAtFrame:
    frame_index: int
    interpolation_type: str
    look_at_point: tuple[float, float, float]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        frame_index = context.stream.readUInt32()
        interpolation_type = context.stream.readFourCc()[::-1]
        if interpolation_type not in ["catm", "line"]:
            raise ValueError(f"Invalid interpolation type: {interpolation_type}")

        look_at = context.stream.readVector3()

        return cls(frame_index=frame_index, interpolation_type=interpolation_type, look_at_point=look_at)


@dataclass
class LookAtCameraAnimationCameraFrame:
    frame_index: int
    interpolation_type: str
    position: tuple[float, float, float]
    roll: float
    fov: float

    @classmethod
    def parse(cls, context: "ParsingContext"):
        frame_index = context.stream.readUInt32()
        interpolation_type = context.stream.readFourCc()[::-1]
        if interpolation_type not in ["catm", "line"]:
            raise ValueError(f"Invalid interpolation type: {interpolation_type}")

        position = context.stream.readVector3()
        roll = context.stream.readFloat()
        fov = context.stream.readFloat()

        return cls(
            frame_index=frame_index,
            interpolation_type=interpolation_type,
            position=position,
            roll=roll,
            fov=fov,
        )


@dataclass
class LookAtCameraAnimationFrameData:
    camera_frames: list[LookAtCameraAnimationCameraFrame]
    look_at_frames: list[LookAtCameraAnimationLookAtFrame]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        camera_frames_count = context.stream.readUInt32()
        camera_frames = []
        for _ in range(camera_frames_count):
            camera_frames.append(LookAtCameraAnimationCameraFrame.parse(context))

        look_at_frames_count = context.stream.readUInt32()
        look_at_frames = []
        for _ in range(look_at_frames_count):
            look_at_frames.append(LookAtCameraAnimationLookAtFrame.parse(context))

        return cls(camera_frames=camera_frames, look_at_frames=look_at_frames)


@dataclass
class CameraAnimation:
    animation_type: str
    name: str
    num_frames: int
    start_offset: int
    frame_data: FreeCameraAnimationFrameData | LookAtCameraAnimationFrameData

    @classmethod
    def parse(cls, context: "ParsingContext"):
        animation_type = context.stream.readFourCc()[::-1]
        if animation_type not in ["free", "look"]:
            raise ValueError(f"Unknown camera animation type: {animation_type}")

        name = context.stream.readUInt16PrefixedAsciiString()
        num_frames = context.stream.readUInt32()
        start_offset = context.stream.readUInt32()

        if animation_type == "free":
            frame_data = FreeCameraAnimationFrameData.parse(context)
        elif animation_type == "look":
            frame_data = LookAtCameraAnimationFrameData.parse(context)
        else:
            raise ValueError(f"Unhandled camera animation type: {animation_type}")

        context.logger.debug(
            f"CameraAnimation: {name}, Type: {animation_type}, Frames: {num_frames}, StartOffset: {start_offset}"
        )
        return cls(
            animation_type=animation_type,
            name=name,
            num_frames=num_frames,
            start_offset=start_offset,
            frame_data=frame_data,
        )


@dataclass
class CameraAnimationList:
    asset_name = "CameraAnimationList"

    version: int
    animations: list[CameraAnimation]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            animations = []
            animation_count = context.stream.readUInt32()
            for _ in range(animation_count):
                animations.append(CameraAnimation.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version=version, animations=animations)



"""Test EnvironmentData asset parsing."""

from sagemap.assets import EnvironmentData
from .conftest import create_context, load_asset_bytes


def test_environment_data():
    """Test EnvironmentData asset parsing."""
    asset_bytes = load_asset_bytes("EnvironmentData")
    
    context = create_context(asset_bytes)
    result = EnvironmentData.parse(context)
    assert result is not None



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class SkippedAsset:
    asset_name = "SkippedAsset"

    version: int
    datasize: int
    data: bytes
    name: str

    @classmethod
    def parse(cls, context: "ParsingContext", name: str):
        version = context.stream.readUInt16()
        datasize = context.stream.readUInt32()
        data = context.stream.readBytes(datasize)

        context.logger.debug(f"Skipped asset: {name}, Version: {version}, Data Size: {datasize} bytes")
        return cls(version, datasize, data, name)



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class AssetListItem:
    type_id: int
    isntance_id: int

    @classmethod
    def parse(cls, context: "ParsingContext"):
        type_id = context.stream.readUInt32()
        instance_id = context.stream.readUInt32()

        context.logger.debug(f"Parsed AssetListItem: Type ID: {type_id}, Instance ID: {instance_id}")
        return cls(type_id, instance_id)


@dataclass
class AssetList:
    asset_name = "AssetList"

    version: int
    asset_names: list[AssetListItem]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            asset_count = context.stream.readUInt32()
            asset_names = [AssetListItem.parse(context) for _ in range(asset_count)]

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, asset_names)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class ScriptArgument:
    type: int
    int_value: int | None = None
    float_value: float | None = None
    string_value: str | None = None
    position_value: tuple[float, float, float] | None = None

    @classmethod
    def parse(cls, context: "ParsingContext"):
        argument_type = context.stream.readUInt32()
        int_value = None
        float_value = None
        string_value = None
        position_value = None

        if argument_type == 16:  # Coordinate
            position_value = (
                context.stream.readFloat(),
                context.stream.readFloat(),
                context.stream.readFloat(),
            )
        else:
            int_value = context.stream.readInt32()
            float_value = context.stream.readFloat()
            string_value = context.stream.readUInt16PrefixedAsciiString()

        return cls(
            type=argument_type,
            int_value=int_value,
            float_value=float_value,
            string_value=string_value,
            position_value=position_value,
        )


@dataclass
class ScriptDerived:
    """Base class for ScriptAction and Condition"""

    content_type: int
    internal_name: str | None
    arguments: list[ScriptArgument]
    is_enabled: bool | None
    is_inverted: bool | None

    @classmethod
    def parse(
        cls,
        context: "ParsingContext",
        has_internal_name_version: int,
        has_is_enabled_version: int,
        has_is_inverted: bool,
    ):
        with context.read_asset() as (asset_version, _):
            content_type = context.stream.readUInt32()

            internal_name = None
            if asset_version >= has_internal_name_version:
                internal_name = context.parse_asset_property_key()

            num_arguments = context.stream.readUInt32()
            arguments = []
            for _ in range(num_arguments):
                arguments.append(ScriptArgument.parse(context))

            is_enabled = None
            is_inverted = None
            if asset_version >= has_is_enabled_version:
                is_enabled = context.stream.readBoolUInt32()

                if has_is_inverted:
                    is_inverted = context.stream.readBoolUInt32()

        context.logger.debug("Finished parsing ScriptDerived")
        return cls(
            content_type=content_type,
            internal_name=internal_name,
            arguments=arguments,
            is_enabled=is_enabled,
            is_inverted=is_inverted,
        )


@dataclass
class OrCondition:
    version: int
    conditions: list[ScriptDerived]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, datasize):
            conditions = []
            condition_end_pos = context.stream.tell() + datasize

            while context.stream.tell() < condition_end_pos:
                asset_name = context.parse_asset_name()
                if asset_name != "Condition":
                    raise ValueError(f"Expected Condition asset, got {asset_name}")

                condition = ScriptDerived.parse(context, 4, 5, True)
                conditions.append(condition)

        context.logger.debug("Finished parsing OrCondition")
        return cls(version=version, conditions=conditions)


@dataclass
class Script:
    name: str
    comment: str
    conditions_comment: str
    actions_comment: str
    is_active: bool
    deactivate_upon_success: bool
    active_in_easy: bool
    active_in_medium: bool
    active_in_hard: bool
    is_subroutine: bool
    evaluation_interval: int | None = None
    uses_evaluation_interval_type: bool = False
    evaluation_interval_type: int = 6
    actions_fire_sequentially: bool | None = None
    loop_actions: bool | None = None
    loop_count: int | None = None
    sequential_target_type: bool | None = None
    sequential_target_name: str | None = None
    unknown: str | None = None
    unknown2: int | None = None
    unknown3: int | None = None
    or_conditions: list[OrCondition] = field(default_factory=list)
    actions_if_true: list[ScriptDerived] = field(default_factory=list)
    actions_if_false: list[ScriptDerived] = field(default_factory=list)

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (asset_version, datasize):
            script_end_pos = context.stream.tell() + datasize

            name = context.stream.readUInt16PrefixedAsciiString()
            context.logger.info(f"Parsing script: {name}")
            comment = context.stream.readUInt16PrefixedAsciiString()
            conditions_comment = context.stream.readUInt16PrefixedAsciiString()
            actions_comment = context.stream.readUInt16PrefixedAsciiString()

            is_active = context.stream.readBool()
            deactivate_upon_success = context.stream.readBool()

            active_in_easy = context.stream.readBool()
            active_in_medium = context.stream.readBool()
            active_in_hard = context.stream.readBool()

            is_subroutine = context.stream.readBool()

            evaluation_interval = None
            uses_evaluation_interval_type = False
            evaluation_interval_type = 6
            if asset_version >= 2:
                evaluation_interval = context.stream.readUInt32()

                if asset_version == 5:
                    uses_evaluation_interval_type = context.stream.readBool()
                    evaluation_interval_type = context.stream.readUInt32()

            actions_fire_sequentially = None
            loop_actions = None
            loop_count = None
            sequential_target_type = None
            sequential_target_name = None
            if asset_version >= 3:
                actions_fire_sequentially = context.stream.readBool()
                loop_actions = context.stream.readBool()
                loop_count = context.stream.readInt32()
                sequential_target_type = context.stream.readBool()
                sequential_target_name = context.stream.readUInt16PrefixedAsciiString()

            unknown = None
            unknown2 = None
            unknown3 = None
            if asset_version >= 4:
                unknown = context.stream.readUInt16PrefixedAsciiString()
                if unknown not in ("ALL", "Planning", "X"):
                    raise ValueError("Invalid data in 'unknown'")

            if asset_version >= 6:
                unknown2 = context.stream.readInt32()
                unknown3 = context.stream.readUInt16()
                if unknown3 != 0:
                    raise ValueError("Invalid data in 'unknown3'")

            or_conditions = []
            actions_if_true = []
            actions_if_false = []
            while context.stream.tell() < script_end_pos:
                asset_name = context.parse_asset_name()
                if asset_name == "OrCondition":
                    or_conditions.append(OrCondition.parse(context))
                elif asset_name == "ScriptAction":
                    actions_if_true.append(ScriptDerived.parse(context, 2, 3, False))
                elif asset_name == "ScriptActionFalse":
                    actions_if_false.append(ScriptDerived.parse(context, 2, 3, False))
                else:
                    raise ValueError(f"Unexpected asset in script: {asset_name}")

        context.logger.debug("Finished parsing Script")
        return cls(
            name=name,
            comment=comment,
            conditions_comment=conditions_comment,
            actions_comment=actions_comment,
            is_active=is_active,
            deactivate_upon_success=deactivate_upon_success,
            active_in_easy=active_in_easy,
            active_in_medium=active_in_medium,
            active_in_hard=active_in_hard,
            is_subroutine=is_subroutine,
            evaluation_interval=evaluation_interval,
            uses_evaluation_interval_type=uses_evaluation_interval_type,
            evaluation_interval_type=evaluation_interval_type,
            actions_fire_sequentially=actions_fire_sequentially,
            loop_actions=loop_actions,
            loop_count=loop_count,
            sequential_target_type=sequential_target_type,
            sequential_target_name=sequential_target_name,
            unknown=unknown,
            unknown2=unknown2,
            unknown3=unknown3,
            or_conditions=or_conditions,
            actions_if_true=actions_if_true,
            actions_if_false=actions_if_false,
        )


@dataclass
class ScriptGroup:
    version: int
    name: str
    is_active: bool
    is_subroutine: bool
    script_groups: dict[str, "ScriptGroup"] = field(default_factory=dict)
    scripts: dict[str, Script] = field(default_factory=dict)

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, datasize):
            group_end_pos = context.stream.tell() + datasize
            name = context.stream.readUInt16PrefixedAsciiString()
            is_active = context.stream.readBool()
            is_subroutine = context.stream.readBool()

            script_groups = {}
            scripts = {}

            while context.stream.tell() < group_end_pos:
                nested_groups, nested_scripts = ScriptGroup.parse_script_list(context)
                script_groups.update(nested_groups)
                scripts.update(nested_scripts)

        context.logger.debug("Finished parsing ScriptGroup")
        return cls(
            version=version,
            name=name,
            is_active=is_active,
            is_subroutine=is_subroutine,
            script_groups=script_groups,
            scripts=scripts,
        )

    @staticmethod
    def parse_script_list(
        context: "ParsingContext",
    ) -> tuple[dict[str, "ScriptGroup"], dict[str, Script]]:
        script_groups = {}
        scripts = {}
        asset_name = context.parse_asset_name()

        if asset_name == "ScriptGroup":
            group = ScriptGroup.parse(context)
            script_groups[group.name] = group
        elif asset_name == "Script":
            script = Script.parse(context)
            scripts[script.name] = script
        else:
            raise ValueError(f"Expected ScriptGroup or Script asset, got {asset_name}")

        return script_groups, scripts


@dataclass
class ScriptList:
    script_groups: dict[str, ScriptGroup]
    scripts: dict[str, Script]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (asset_version, datasize):
            script_list_end_pos = context.stream.tell() + datasize

            if asset_version != 1:
                raise ValueError(f"Unexpected ScriptList version: {asset_version}")

            script_groups = {}
            scripts = {}

            while context.stream.tell() < script_list_end_pos:
                nested_groups, nested_scripts = ScriptGroup.parse_script_list(context)
                script_groups.update(nested_groups)
                scripts.update(nested_scripts)

        context.logger.debug("Finished parsing ScriptList")
        return cls(script_groups=script_groups, scripts=scripts)


@dataclass
class PlayerScriptsList:
    asset_name = "PlayerScriptsList"

    version: int
    script_lists: list[ScriptList]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, datasize):
            player_script_list_end_pos = context.stream.tell() + datasize

            script_lists = []
            while context.stream.tell() < player_script_list_end_pos:
                asset_name = context.parse_asset_name()
                if asset_name != "ScriptList":
                    raise ValueError(f"Expected ScriptList asset, got {asset_name}")

                script_lists.append(ScriptList.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version=version, script_lists=script_lists)



"""Test AssetList asset parsing."""

import pytest
from sagemap.assets import AssetList
from .conftest import create_context, load_asset_bytes

@pytest.mark.skip(reason="No test data available for AssetList yet")
def test_asset_list():
    """Test AssetList asset parsing."""
    asset_bytes = load_asset_bytes("AssetList")
    
    context = create_context(asset_bytes)
    result = AssetList.parse(context)
    assert result is not None



"""Test Teams asset parsing."""

from sagemap.assets import Teams
from .conftest import create_context, load_asset_bytes


def test_teams():
    """Test Teams asset parsing."""
    asset_bytes = load_asset_bytes("Teams")
    
    context = create_context(asset_bytes)
    result = Teams.parse(context)
    assert result is not None
    assert hasattr(result, 'teams')
    assert hasattr(result, 'version')



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class HeightMapData:
    asset_name = "HeightMapData"

    version: int
    width: int
    height: int
    border_width: int
    borders: list[dict[str, tuple[int, int]]]
    area: int
    min_height: int
    max_height: int
    elevations: list[list[int]]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            width = context.stream.readUInt32()
            height = context.stream.readUInt32()
            border_width = context.stream.readUInt32()

            border_count = context.stream.readUInt32()
            borders = []
            for _ in range(border_count):
                if version >= 6:
                    corner1X = context.stream.readUInt32()
                    corner1Y = context.stream.readUInt32()
                    x = context.stream.readUInt32()
                    y = context.stream.readUInt32()
                else:
                    corner1X = 0
                    corner1Y = 0
                    x = context.stream.readUInt32()
                    y = context.stream.readUInt32()

                borders.append({"corner1": (corner1X, corner1Y), "position": (x, y)})

            area = context.stream.readUInt32()
            if area != width * height:
                raise ValueError(f"Invalid area: {area}, expected: {width * height}")

            min_height = None
            max_height = None
            elevations = [[0 for _ in range(width)] for _ in range(height)]
            for y in range(height):
                for x in range(width):
                    if version >= 5:
                        elevation = context.stream.readUInt16()
                    else:
                        elevation = context.stream.readUChar()
                    elevations[y][x] = elevation
                    if min_height is None or elevation < min_height:
                        min_height = elevation
                    if max_height is None or elevation > max_height:
                        max_height = elevation

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, width, height, border_width, borders, area, min_height, max_height, elevations)



from dataclasses import dataclass

from sagemap.context import ParsingContext


@dataclass
class WaypointsList:
    asset_name = "WaypointsList"

    version: int
    waypoint_paths: list[tuple[int, int]]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            waypoint_paths = []
            waypoint_count = context.stream.readUInt32()
            for _ in range(waypoint_count):
                start_waypoint_id = context.stream.readUInt32()
                end_waypoint_id = context.stream.readUInt32()

                waypoint_paths.append((start_waypoint_id, end_waypoint_id))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, waypoint_paths)



"""Test BlendTileData asset parsing."""

from sagemap.assets import BlendTileData, HeightMapData
from .conftest import create_context, load_asset_bytes


def test_blend_tile_data():
    """Test BlendTileData asset parsing."""
    asset_bytes = load_asset_bytes("BlendTileData")
    heght_map_bytes = load_asset_bytes("HeightMapData")
    
    context = create_context(asset_bytes)
    height_map_context = create_context(heght_map_bytes)

    height_map = HeightMapData.parse(height_map_context)
    result = BlendTileData.parse(context, height_map)
    assert result is not None



import io
import struct


class BinaryStream:
    def __init__(self, base_stream, encoding="latin-1"):
        self.base_stream: io.BytesIO = base_stream
        self.encoding = encoding

    def readByte(self) -> bytes:
        return self.base_stream.read(1)

    def writeByte(self, value: int):
        self.base_stream.write(bytes([value]))

    def readBytes(self, length) -> bytes:
        return self.base_stream.read(length)

    def writeBytes(self, value: bytes):
        self.base_stream.write(value)

    def readChar(self) -> int:
        return self.unpack("b")

    def writeChar(self, value: bytes):
        self.pack("c", value)

    def readUChar(self) -> int:
        return self.unpack("B")

    def writeUChar(self, value: int):
        self.pack("C", value)

    def readBool(self) -> bool:
        return self.unpack("?")

    def writeBool(self, value: bool):
        self.pack("?", value)

    def readBoolUInt32(self) -> bool:
        result = self.readBool()
        unused = self.readUInt24()  # padding

        if unused != 0:
            raise ValueError("Expected padding bytes to be zero")

        return result

    def writeBoolUInt32(self, value: bool):
        self.writeBool(value)
        self.writeUInt24(0)

    def readInt16(self) -> int:
        return self.unpack("h", 2)

    def writeInt16(self, value: int):
        self.pack("h", value)

    def readUInt16(self) -> int:
        return self.unpack("H", 2)

    def writeUInt16(self, value: int):
        self.pack("H", value)

    def readInt32(self) -> int:
        return self.unpack("i", 4)

    def writeInt32(self, value: int):
        self.pack("i", value)

    def readUInt32(self) -> int:
        return self.unpack("I", 4)

    def writeUInt32(self, value: int):
        self.pack("I", value)

    def readInt64(self) -> int:
        return self.unpack("q", 8)

    def writeInt64(self, value: int):
        self.pack("q", value)

    def readUInt64(self) -> int:
        return self.unpack("Q", 8)

    def writeUInt64(self, value: int):
        self.pack("Q", value)

    def readFloat(self) -> float:
        return self.unpack("f", 4)

    def writeFloat(self, value: float):
        self.pack("f", value)

    def readDouble(self) -> float:
        return self.unpack("d", 8)

    def writeDouble(self, value: float):
        self.pack("d", value)

    def readVector2(self) -> tuple[float, float]:
        return (self.readFloat(), self.readFloat())

    def readVector3(self) -> tuple[float, float, float]:
        return (self.readFloat(), self.readFloat(), self.readFloat())

    def readVector4(self) -> tuple[float, float, float, float]:
        return (self.readFloat(), self.readFloat(), self.readFloat(), self.readFloat())

    def readString(self) -> str:
        length = self.readUChar()
        return self.unpack(str(length) + "s", length).decode(self.encoding)

    def writeString(self, value: str):
        length = len(value)
        self.writeUChar(length)
        self.pack(str(length) + "s", value.encode(self.encoding))

    def readUInt16PrefixedAsciiString(self) -> str:
        lenght = self.readUInt16()
        return self.readBytes(lenght).decode(self.encoding)

    def writeUInt16PrefixedAsciiString(self, value: str):
        lenght = len(value)
        self.writeUInt16(lenght)
        self.writeBytes(value.encode(self.encoding))

    def readFourCc(self) -> str:
        return self.readBytes(4).decode(self.encoding)

    def writeFourCc(self, value: str):
        if len(value) != 4:
            raise ValueError("FourCC must be exactly 4 characters")
        self.writeBytes(value.encode(self.encoding))

    def readUInt24(self) -> int:
        b = self.readBytes(3)
        return int.from_bytes(b, "little", signed=False)

    def writeUInt24(self, value: int):
        if not (0 <= value <= 0xFFFFFF):
            raise ValueError("Value out of range for UInt24 (0..16777215)")
        b = value.to_bytes(3, "little")
        self.writeBytes(b)

    def readUInt16PrefixedUnicodeString(self) -> str:
        length = self.readUInt16()
        return self.readBytes(length * 2).decode("utf-16-le")

    def writeUInt16PrefixedUnicodeString(self, value: str):
        encoded = value.encode("utf-16-le")
        self.writeUInt16(len(value))
        self.writeBytes(encoded)

    def readUInt16Array2D(self, width: int, height: int) -> list[list[int]]:
        result = [[0] * height for _ in range(width)]
        for y in range(height):
            for x in range(width):
                result[x][y] = self.readUInt16()
        return result

    def writeUInt16Array2D(self, array2d: list[list[int]], width: int, height: int):
        for y in range(height):
            for x in range(width):
                self.writeUInt16(array2d[x][y])

    def readUIntArray2D(self, width: int, height: int, bit_size: int) -> list[list[int]]:
        """Read a 2D array of unsigned integers.

        Args:
            width: Width of the array
            height: Height of the array
            bit_size: Size in bits (16 or 32) - determines whether to read UInt16 or UInt32
        """
        result = [[0] * height for _ in range(width)]

        for y in range(height):
            for x in range(width):
                if bit_size == 16:
                    value = self.readUInt16()
                elif bit_size == 32:
                    value = self.readUInt32()
                else:
                    raise ValueError(f"Unsupported bit_size: {bit_size}. Expected 16 or 32.")

                result[x][y] = value

        return result

    def readSingleBitBooleanArray2D(self, width: int, height: int, row_byte_aligned: bool = True) -> list[list[bool]]:
        """Read a 2D array of single-bit boolean values.

        Args:
            width: Width of the array
            height: Height of the array
            row_byte_aligned: If True (default), each row starts on a byte boundary matching C# behavior.
                             If False, bits flow continuously (non-standard).
        """
        result = [[False] * height for _ in range(width)]

        if row_byte_aligned:
            # Each row starts on a fresh byte boundary
            for y in range(height):
                temp = 0
                for x in range(width):
                    if x % 8 == 0:
                        temp = self.readUChar()

                    result[x][y] = (temp & (1 << (x % 8))) != 0
        else:
            # Bits flow continuously without row alignment
            temp = 0
            bit_index = 0
            for y in range(height):
                for x in range(width):
                    if bit_index % 8 == 0:
                        temp = self.readUChar()

                    result[x][y] = (temp & (1 << (bit_index % 8))) != 0
                    bit_index += 1

            # Note: If we ended mid-byte, remaining bits are padding
            # This is implicit since bits_read will be reset on next read

        return result

    def readByteArray2D(self, width: int, height: int) -> list[list[int]]:
        result = [[0] * height for _ in range(width)]
        for y in range(height):
            for x in range(width):
                result[x][y] = self.readUChar()
        return result

    def readByteArray2DAsEnum(self, width: int, height: int, enum_class):
        result = [[None] * height for _ in range(width)]
        for y in range(height):
            for x in range(width):
                result[x][y] = enum_class(self.readUChar())
        return result

    def readSingle(self):
        return self.unpack("f", 4)

    def writeSingle(self, value: float):
        self.pack("f", value)

    def pack(self, fmt, data):
        return self.writeBytes(struct.pack(fmt, data))

    def unpack(self, fmt, length=1):
        return struct.unpack(fmt, self.readBytes(length))[0]

    def seek(self, offset, whence=io.SEEK_SET):
        self.base_stream.seek(offset, whence)

    def tell(self):
        return self.base_stream.tell()

    def getvalue(self):
        return self.base_stream.getvalue()



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class MPPosition:
    asset_name = "MPPosition"

    version: int
    is_human: bool
    is_computer: bool
    load_ai_script: bool
    team: int
    side_restrictions: list[str]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            is_human = context.stream.readBool()
            is_computer = context.stream.readBool()
            load_ai_script = False
            if version > 0:
                load_ai_script = context.stream.readBool()

            team = context.stream.readUInt32()
            side_restrictions = []
            if version > 0:
                side_restriction_count = context.stream.readUInt32()
                for _ in range(side_restriction_count):
                    side_restrictions.append(context.stream.readUInt16PrefixedAsciiString())

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, is_human, is_computer, load_ai_script, team, side_restrictions)


@dataclass
class MPPositionsList:
    asset_name = "MPPositionList"

    version: int
    positions: list[MPPosition]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, datasize):
            positions = []
            end_pos = context.stream.tell() + datasize
            while context.stream.tell() < end_pos:
                name = context.parse_asset_name()
                if name != "MPPositionInfo":
                    raise ValueError(f"Expected MPPositionInfo asset, got {name}")

                positions.append(MPPosition.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, positions)



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class EnvironmentData:
    asset_name = "EnvironmentData"

    version: int
    water_max_alpha_depth: float | None
    deep_water_alpha: float | None
    is_macro_texture_stretched: bool | None
    macro_texture: str
    cloud_texture: str
    unknown_texture: str | None
    unknown_texture2: str | None

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, datasize):
            end_pos = context.stream.tell() + datasize

            water_max_alpha_depth = None
            deep_water_alpha = None
            if version >= 3:
                water_max_alpha_depth = context.stream.readFloat()
                deep_water_alpha = context.stream.readFloat()

            is_macro_texture_stretched = None
            if version < 5:
                is_macro_texture_stretched = context.stream.readBool()

            macro_texture = context.stream.readUInt16PrefixedAsciiString()
            cloud_texture = context.stream.readUInt16PrefixedAsciiString()

            unknown_texture = None
            if version >= 4:
                unknown_texture = context.stream.readUInt16PrefixedAsciiString()

            unknown_texture2 = None
            if version >= 6 and context.stream.tell() < end_pos:
                unknown_texture2 = context.stream.readUInt16PrefixedAsciiString()

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(
            version,
            water_max_alpha_depth,
            deep_water_alpha,
            is_macro_texture_stretched,
            macro_texture,
            cloud_texture,
            unknown_texture,
            unknown_texture2,
        )



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class StandingWaterArea:
    asset_name = "StandingWaterArea"

    unique_id: int
    name: str
    layer_name: str | None
    uv_scroll_speed: float
    use_adaptive_blending: bool
    bump_map_texture: str
    sky_texture: str
    points: list[tuple[float, float]]
    water_height: int
    fx_shader: str
    depth_color: str

    @classmethod
    def parse(cls, context: "ParsingContext"):
        unique_id = context.stream.readUInt32()
        name = context.stream.readUInt16PrefixedAsciiString()
        layer_name = context.stream.readUInt16PrefixedAsciiString()
        uv_scroll_speed = context.stream.readSingle()
        use_adaptive_blending = context.stream.readBool()
        bump_map_texture = context.stream.readUInt16PrefixedAsciiString()
        sky_texture = context.stream.readUInt16PrefixedAsciiString()

        point_count = context.stream.readUInt32()
        points = []
        for _ in range(point_count):
            points.append(context.stream.readVector2())

        water_height = context.stream.readUInt32()
        fx_shader = context.stream.readUInt16PrefixedAsciiString()
        depth_color = context.stream.readUInt16PrefixedAsciiString()

        return cls(
            unique_id=unique_id,
            name=name,
            layer_name=layer_name,
            uv_scroll_speed=uv_scroll_speed,
            use_adaptive_blending=use_adaptive_blending,
            bump_map_texture=bump_map_texture,
            sky_texture=sky_texture,
            points=points,
            water_height=water_height,
            fx_shader=fx_shader,
            depth_color=depth_color,
        )


@dataclass
class StandingWaterAreas:
    asset_name = "StandingWaterAreas"

    version: int
    areas: list[StandingWaterArea]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            area_count = context.stream.readUInt32()
            areas = []

            for _ in range(area_count):
                areas.append(StandingWaterArea.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, areas)



import base64
import io
import logging
from dataclasses import asdict, is_dataclass
from enum import Enum

from reversebox.compression.compression_refpack import RefpackHandler

from .assets import (
    AssetList,
    BlendTileData,
    BuildLists,
    CameraAnimationList,
    EnvironmentData,
    GlobalLighting,
    GlobalVersion,
    HeightMapData,
    LibraryMapLists,
    MPPositionsList,
    NamedCameras,
    ObjectsList,
    PlayerScriptsList,
    PolygonTriggers,
    PostEffectsChunk,
    RiverAreas,
    SidesList,
    SkippedAsset,
    StandingWaterAreas,
    StandingWaveAreas,
    Teams,
    TriggerAreas,
    WaterSettings,
    WaypointsList,
    WorldInfo,
)
from .context import ParsingContext
from .stream import BinaryStream


class Map:
    global_version: GlobalVersion
    height_map_data: HeightMapData
    blend_tile_data: BlendTileData
    world_info: WorldInfo
    objects_list: ObjectsList
    waypoints_list: WaypointsList
    sides_list: SidesList
    player_scripts_list: PlayerScriptsList
    water_settings: WaterSettings
    asset_list: AssetList
    build_lists: BuildLists
    polygon_triggers: PolygonTriggers
    trigger_areas: TriggerAreas
    standing_water_areas: StandingWaterAreas
    standing_wave_areas: StandingWaveAreas
    river_areas: RiverAreas
    global_lighting: GlobalLighting
    environment_data: EnvironmentData
    post_effects_chunk: PostEffectsChunk
    named_cameras: NamedCameras
    camera_animation_list: CameraAnimationList
    library_map_lists: LibraryMapLists
    teams: Teams
    mp_positions_list: MPPositionsList

    skipped_assets: dict[str, SkippedAsset]

    def __init__(self, context: ParsingContext = None):
        self.context = context

        self.compression_bytes = None
        self.asset_count = None
        self.assets = {}

        # assets
        self.global_version = None
        self.height_map_data = None
        self.blend_tile_data = None
        self.world_info = None
        self.objects_list = None
        self.waypoints_list = None
        self.sides_list = None
        self.player_scripts_list = None
        self.water_settings = None
        self.asset_list = None
        self.build_lists = None
        self.polygon_triggers = None
        self.trigger_areas = None
        self.standing_water_areas = None
        self.standing_wave_areas = None
        self.river_areas = None
        self.global_lighting = None
        self.environment_data = None
        self.post_effects_chunk = None
        self.named_cameras = None
        self.camera_animation_list = None
        self.library_map_lists = None

        self.skipped_assets = {}

    def parse(self):
        self.assets = self.context.parse_assets()
        while self.context.stream.tell() < len(self.context.stream.getvalue()):
            asset_name = self.context.parse_asset_name()
            self.context.logger.info(f"Processing asset: {asset_name}")
            self.parse_asset(asset_name)

    def parse_asset(self, asset_name: str):
        if asset_name == AssetList.asset_name:
            self.asset_list = AssetList.parse(self.context)
        elif asset_name == HeightMapData.asset_name:
            self.height_map_data = HeightMapData.parse(self.context)
        elif asset_name == WorldInfo.asset_name:
            self.world_info = WorldInfo.parse(self.context)
        elif asset_name == Teams.asset_name:
            self.teams = Teams.parse(self.context)
        elif asset_name == PlayerScriptsList.asset_name:
            self.player_scripts_list = PlayerScriptsList.parse(self.context)
        elif asset_name == ObjectsList.asset_name:
            self.objects_list = ObjectsList.parse(self.context)
        elif asset_name == GlobalVersion.asset_name:
            self.global_version = GlobalVersion.parse(self.context)
        elif asset_name == BlendTileData.asset_name:
            self.blend_tile_data = BlendTileData.parse(self.context, self.height_map_data)
        elif asset_name == MPPositionsList.asset_name:
            self.mp_positions_list = MPPositionsList.parse(self.context)
        elif asset_name == SidesList.asset_name:
            self.sides_list = SidesList.parse(self.context, self.asset_list is not None)
        elif asset_name == TriggerAreas.asset_name:
            self.trigger_areas = TriggerAreas.parse(self.context)
        elif asset_name == PolygonTriggers.asset_name:
            self.polygon_triggers = PolygonTriggers.parse(self.context)
        elif asset_name == WaypointsList.asset_name:
            self.waypoints_list = WaypointsList.parse(self.context)
        elif asset_name == WaterSettings.asset_name:
            self.water_settings = WaterSettings.parse(self.context)
        elif asset_name == BuildLists.asset_name:
            self.build_lists = BuildLists.parse(self.context, self.asset_list is not None)
        elif asset_name == StandingWaterAreas.asset_name:
            self.standing_water_areas = StandingWaterAreas.parse(self.context)
        elif asset_name == StandingWaveAreas.asset_name:
            self.standing_wave_areas = StandingWaveAreas.parse(self.context)
        elif asset_name == RiverAreas.asset_name:
            self.river_areas = RiverAreas.parse(self.context)
        elif asset_name == GlobalLighting.asset_name:
            self.global_lighting = GlobalLighting.parse(self.context)
        elif asset_name == EnvironmentData.asset_name:
            self.environment_data = EnvironmentData.parse(self.context)
        elif asset_name == PostEffectsChunk.asset_name:
            self.post_effects_chunk = PostEffectsChunk.parse(self.context)
        elif asset_name == NamedCameras.asset_name:
            self.named_cameras = NamedCameras.parse(self.context)
        elif asset_name == CameraAnimationList.asset_name:
            self.camera_animation_list = CameraAnimationList.parse(self.context)
        elif asset_name == LibraryMapLists.asset_name:
            self.library_map_lists = LibraryMapLists.parse(self.context)
        else:
            self.context.logger.debug(f"Unknown asset: {asset_name}, skipping")
            self.skipped_assets[asset_name] = SkippedAsset.parse(self.context, asset_name)

    def to_dict(self):
        """Convert Map and all assets to a JSON-serializable dictionary"""
        result = {}

        for key, value in self.__dict__.items():
            if key == "context":  # Skip the parsing context
                continue
            result[key] = self._serialize(value)

        return result

    def _serialize(self, obj):
        """Recursively serialize objects to JSON-compatible types"""
        if obj is None:
            return None
        elif isinstance(obj, bytes):
            return base64.b64encode(obj).decode("ascii")
        elif isinstance(obj, Enum):
            return obj.value
        elif is_dataclass(obj):
            return {k: self._serialize(v) for k, v in asdict(obj).items()}
        elif isinstance(obj, dict):
            return {(k.name if isinstance(k, Enum) else k): self._serialize(v) for k, v in obj.items()}
        elif isinstance(obj, (list, tuple)):
            return [self._serialize(item) for item in obj]
        else:
            return obj


def parse_map(file: io.BufferedReader) -> Map:
    ea_compression = file.read(8)
    if not ea_compression.startswith(b"EAR"):
        file.seek(0)

    compressed_data = file.read()

    try:
        decompressed_data = RefpackHandler().decompress_data(compressed_data)
    except Exception:
        decompressed_data = compressed_data

    logger = logging.getLogger("sagemap")

    stream = BinaryStream(io.BytesIO(decompressed_data))
    context = ParsingContext(stream)
    context.set_logger(logger)

    map = Map(context)
    map.parse()

    return map


def parse_map_from_path(path: str) -> Map:
    with open(path, "rb") as file:
        return parse_map(file)



"""Test SidesList and BuildLists asset parsing."""

from sagemap.assets import SidesList
from .conftest import create_context, load_asset_bytes


def test_sides_list():
    """Test SidesList asset parsing."""
    asset_bytes = load_asset_bytes("SidesList")
    
    context = create_context(asset_bytes)
    result = SidesList.parse(context, False)
    assert result is not None




"""Test RiverAreas asset parsing."""

from sagemap.assets import RiverAreas
from .conftest import create_context, load_asset_bytes


def test_river_areas():
    """Test RiverAreas asset parsing."""
    asset_bytes = load_asset_bytes("RiverAreas")
    
    context = create_context(asset_bytes)
    result = RiverAreas.parse(context)
    assert result is not None



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class StandingWaveArea:
    asset_name = "StandingWaveArea"

    unique_id: int
    name: str
    layer_name: str | None
    uv_scroll_speed: float
    use_adaptive_blending: bool
    points: list[tuple[float, float]]
    final_width: int | None
    final_height: int | None
    initial_width_fraction: int | None
    initial_height_fraction: int | None
    initial_velocity: int | None
    time_to_fade: int | None
    time_to_compress: int | None
    time_offset_2nd_wave: int | None
    distance_from_shore: int | None
    texture: str | None
    enable_pca_wave: bool | None
    wave_particle_fx_name: str | None

    @classmethod
    def parse(cls, context: "ParsingContext", version: int):
        unique_id = context.stream.readUInt32()
        name = context.stream.readUInt16PrefixedAsciiString()
        layer_name = context.stream.readUInt16PrefixedAsciiString()
        uv_scroll_speed = context.stream.readFloat()
        use_adaptive_blending = context.stream.readBool()

        point_count = context.stream.readUInt32()
        points = []
        for _ in range(point_count):
            points.append(context.stream.readVector2())

        unknown = context.stream.readUInt32()
        if unknown != 0:
            raise ValueError(f"Expected unknown field to be 0, got {unknown}")

        final_width = None
        final_height = None
        initial_width_fraction = None
        initial_height_fraction = None
        initial_velocity = None
        time_to_fade = None
        time_to_compress = None
        time_offset_2nd_wave = None
        distance_from_shore = None
        texture = None
        if version < 3:
            final_width = context.stream.readUInt32()
            final_height = context.stream.readUInt32()
            initial_width_fraction = context.stream.readUInt32()
            initial_height_fraction = context.stream.readUInt32()
            initial_velocity = context.stream.readUInt32()
            time_to_fade = context.stream.readUInt32()
            time_to_compress = context.stream.readUInt32()
            time_offset_2nd_wave = context.stream.readUInt32()
            distance_from_shore = context.stream.readUInt32()
            texture = context.stream.readUInt16PrefixedAsciiString()

        enable_pca_wave = None
        if version == 2:
            enable_pca_wave = context.stream.readBoolUInt32()

        wave_particle_fx_name = None
        if version >= 4:
            wave_particle_fx_name = context.stream.readUInt16PrefixedAsciiString()

        return cls(
            unique_id=unique_id,
            name=name,
            layer_name=layer_name,
            uv_scroll_speed=uv_scroll_speed,
            use_adaptive_blending=use_adaptive_blending,
            points=points,
            final_width=final_width,
            final_height=final_height,
            initial_width_fraction=initial_width_fraction,
            initial_height_fraction=initial_height_fraction,
            initial_velocity=initial_velocity,
            time_to_fade=time_to_fade,
            time_to_compress=time_to_compress,
            time_offset_2nd_wave=time_offset_2nd_wave,
            distance_from_shore=distance_from_shore,
            texture=texture,
            enable_pca_wave=enable_pca_wave,
            wave_particle_fx_name=wave_particle_fx_name,
        )


@dataclass
class StandingWaveAreas:
    asset_name = "StandingWaveAreas"

    version: int
    areas: list[StandingWaveArea]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            area_count = context.stream.readUInt32()
            areas = []

            for _ in range(area_count):
                areas.append(StandingWaveArea.parse(context, version))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version=version, areas=areas)



import re

from setuptools import find_packages, setup

version = ""
with open("sagemap/__init__.py") as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError("version is not set")

readme = ""
with open("README.md") as f:
    readme = f.read()

setup(
    name="sagemap",
    version=version,
    url="https://github.com/ClementJ18/sagemap",
    packages=find_packages(include=["sagemap", "sagemap.*"]),
    description="A library for reading and writing .map files from SAGE engine games.",
    long_description_content_type="text/markdown",
    long_description=readme,
    python_requires=">=3.8",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
)



"""Test GlobalLighting asset parsing."""

from sagemap.assets import GlobalLighting
from .conftest import create_context, load_asset_bytes


def test_global_lighting():
    """Test GlobalLighting asset parsing."""
    asset_bytes = load_asset_bytes("GlobalLighting")
    
    context = create_context(asset_bytes)
    result = GlobalLighting.parse(context)
    assert result is not None



from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext
    from .height_map import HeightMapData


class TileFlammability(IntEnum):
    """Enum for tile flammability values."""

    FIRE_RESISTANT = 0
    GRASS = 1
    HIGHLY_FLAMMABLE = 2
    UNDEFINED = 3


@dataclass
class BlendTileTexture:
    """Represents a blend tile texture (inline data structure, not an asset)."""

    cell_start: int
    cell_count: int
    cell_size: int
    magic_value: int
    name: str

    @classmethod
    def parse(cls, context: "ParsingContext"):
        """Parse inline (no asset header)."""
        cell_start = context.stream.readUInt32()
        cell_count = context.stream.readUInt32()
        cell_size = context.stream.readUInt32()

        if cell_size * cell_size != cell_count:
            raise ValueError(f"Invalid cell_size: {cell_size}^2 != {cell_count}")

        magic_value = context.stream.readUInt32()
        if magic_value != 0:
            raise ValueError(f"Expected magic_value to be 0, got: {magic_value}")

        name = context.stream.readUInt16PrefixedAsciiString()

        return cls(
            cell_start=cell_start,
            cell_count=cell_count,
            cell_size=cell_size,
            magic_value=magic_value,
            name=name,
        )


@dataclass
class BlendDescription:
    """Represents a blend description (inline data structure, not an asset)."""

    secondary_texture_tile: int
    raw_blend_direction: bytes
    flags: int
    two_sided: bool
    magic_value1: int

    @classmethod
    def parse(cls, context: "ParsingContext", version: int):
        """Parse inline (no asset header)."""
        secondary_texture_tile = context.stream.readUInt32()
        raw_blend_direction = context.stream.readBytes(4)
        flags = context.stream.readUChar()
        two_sided = context.stream.readBool()

        magic_value1 = context.stream.readUInt32()
        # MagicValue1 can be 0xFFFFFFFF or 24

        magic_value2 = context.stream.readUInt32()
        if magic_value2 != 0x7ADA0000:
            raise ValueError(f"Expected magic_value2 to be 0x7ADA0000, got: {magic_value2:#x}")

        return cls(
            secondary_texture_tile=secondary_texture_tile,
            raw_blend_direction=raw_blend_direction,
            flags=flags,
            two_sided=two_sided,
            magic_value1=magic_value1,
        )


@dataclass
class CliffTextureMapping:
    """Represents a cliff texture mapping (inline data structure, not an asset)."""

    texture_tile: int
    bottom_left_coords: tuple[float, float]
    bottom_right_coords: tuple[float, float]
    top_right_coords: tuple[float, float]
    top_left_coords: tuple[float, float]
    unknown2: int

    @classmethod
    def parse(cls, context: "ParsingContext"):
        """Parse inline (no asset header)."""
        texture_tile = context.stream.readUInt32()

        # Read 4 Vector2s (each is 2 floats)
        bottom_left_coords = context.stream.readVector2()
        bottom_right_coords = context.stream.readVector2()
        top_right_coords = context.stream.readVector2()
        top_left_coords = context.stream.readVector2()

        unknown2 = context.stream.readUInt16()

        return cls(
            texture_tile=texture_tile,
            bottom_left_coords=bottom_left_coords,
            bottom_right_coords=bottom_right_coords,
            top_right_coords=top_right_coords,
            top_left_coords=top_left_coords,
            unknown2=unknown2,
        )


def get_blend_bit_size(version: int) -> int:
    """Get the blend bit size based on version.
    Returns 32 for versions 14-23, otherwise 16.
    """
    if version >= 14 and version < 24:
        return 32
    else:
        return 16


@dataclass
class BlendTileData:
    asset_name = "BlendTileData"

    tiles: list[list[int]]
    blends: list[list[int]]
    three_way_blends: list[list[int]]
    cliff_textures: list[list[int]]
    impassability: list[list[bool]] | None
    impassability_to_players: list[list[bool]] | None
    passage_widths: list[list[bool]] | None
    taintability: list[list[bool]] | None
    extra_passability: list[list[bool]] | None
    flammability: list[list[TileFlammability]] | None
    visibility: list[list[bool]] | None
    buildability: list[list[bool]] | None
    impassability_to_air_units: list[list[bool]] | None
    tiberium_growability: list[list[bool]] | None
    dynamic_shrubbery_density: list[list[int]] | None
    texture_cell_count: int
    parsed_cliff_texture_mappings_count: int
    textures: list[BlendTileTexture]
    magic_value1: int
    magic_value2: int
    blend_descriptions: list[BlendDescription]
    cliff_texture_mappings: list[CliffTextureMapping]

    @classmethod
    def parse(cls, context: "ParsingContext", height_map_data: "HeightMapData"):
        with context.read_asset() as (version, datasize):
            start_pos = context.stream.tell()
            end_pos = start_pos + datasize

            if version < 6:
                raise ValueError(f"Unsupported BlendTileData version: {version}")

            if height_map_data is None:
                raise ValueError("Expected HeightMapData asset before BlendTileData asset.")

            width = height_map_data.width
            height = height_map_data.height

            tiles_count = context.stream.readUInt32()
            if tiles_count != width * height:
                raise ValueError(f"Invalid tiles_count: {tiles_count}, expected: {width * height}")

            tiles = context.stream.readUInt16Array2D(width, height)

            blend_bit_size = get_blend_bit_size(version)
            blends = context.stream.readUIntArray2D(width, height, blend_bit_size)
            three_way_blends = context.stream.readUIntArray2D(width, height, blend_bit_size)
            cliff_textures = context.stream.readUIntArray2D(width, height, blend_bit_size)

            impassability = None
            if version > 6:
                passability_width = height_map_data.width
                if version == 7:
                    # C&C Generals clips partial bytes from each row of passability data
                    passability_width = ((passability_width + 1) // 8) * 8

                # If terrain is passable, there's a 0 in the data file.
                impassability = context.stream.readSingleBitBooleanArray2D(passability_width, height_map_data.height)

            impassability_to_players = None
            if version >= 10:
                impassability_to_players = context.stream.readSingleBitBooleanArray2D(
                    height_map_data.width, height_map_data.height
                )

            passage_widths = None
            if version >= 11:
                passage_widths = context.stream.readSingleBitBooleanArray2D(
                    height_map_data.width, height_map_data.height
                )

            taintability = None
            if version >= 14 and version < 25:
                taintability = context.stream.readSingleBitBooleanArray2D(height_map_data.width, height_map_data.height)

            extra_passability = None
            if version >= 15:
                extra_passability = context.stream.readSingleBitBooleanArray2D(
                    height_map_data.width, height_map_data.height
                )

            flammability = None
            if version >= 16 and version < 25:
                flammability = context.stream.readByteArray2DAsEnum(
                    height_map_data.width, height_map_data.height, TileFlammability
                )

            visibility = None
            if version >= 17:
                # Note: All ReadSingleBitBooleanArray2D calls in C# use row-byte-alignment
                visibility = context.stream.readSingleBitBooleanArray2D(height_map_data.width, height_map_data.height)

            buildability = None
            impassability_to_air_units = None
            tiberium_growability = None
            if version >= 24:
                # TODO: Are these in the right order?
                buildability = context.stream.readSingleBitBooleanArray2D(height_map_data.width, height_map_data.height)
                impassability_to_air_units = context.stream.readSingleBitBooleanArray2D(
                    height_map_data.width, height_map_data.height
                )
                tiberium_growability = context.stream.readSingleBitBooleanArray2D(
                    height_map_data.width, height_map_data.height
                )

            dynamic_shrubbery_density = None
            if version >= 25:
                dynamic_shrubbery_density = context.stream.readByteArray2D(
                    height_map_data.width, height_map_data.height
                )

            texture_cell_count = context.stream.readUInt32()

            blends_count_raw = context.stream.readUInt32()
            blends_count = blends_count_raw
            if blends_count > 0:
                # Usually minimum value is 1, but some files (perhaps Generals, not Zero Hour?) have 0.
                blends_count -= 1

            parsed_cliff_texture_mappings_count = context.stream.readUInt32()
            cliff_blends_count = parsed_cliff_texture_mappings_count
            if cliff_blends_count > 0:
                # Usually minimum value is 1, but some files (perhaps Generals, not Zero Hour?) have 0.
                cliff_blends_count -= 1

            texture_count = context.stream.readUInt32()
            textures = []
            for i in range(texture_count):
                textures.append(BlendTileTexture.parse(context))

            # Can be a variety of values, don't know what it means.
            magic_value1 = context.stream.readUInt32()

            magic_value2 = context.stream.readUInt32()
            if magic_value2 != 0:
                raise ValueError(f"Expected magic_value2 to be 0, got: {magic_value2}")

            blend_descriptions = []
            for _ in range(blends_count):
                blend_descriptions.append(BlendDescription.parse(context, version))

            cliff_texture_mappings = []
            for _ in range(cliff_blends_count):
                cliff_texture_mappings.append(CliffTextureMapping.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(
            tiles=tiles,
            blends=blends,
            three_way_blends=three_way_blends,
            cliff_textures=cliff_textures,
            impassability=impassability,
            impassability_to_players=impassability_to_players,
            passage_widths=passage_widths,
            taintability=taintability,
            extra_passability=extra_passability,
            flammability=flammability,
            visibility=visibility,
            buildability=buildability,
            impassability_to_air_units=impassability_to_air_units,
            tiberium_growability=tiberium_growability,
            dynamic_shrubbery_density=dynamic_shrubbery_density,
            texture_cell_count=texture_cell_count,
            parsed_cliff_texture_mappings_count=parsed_cliff_texture_mappings_count,
            textures=textures,
            magic_value1=magic_value1,
            magic_value2=magic_value2,
            blend_descriptions=blend_descriptions,
            cliff_texture_mappings=cliff_texture_mappings,
        )



"""Test MPPositionsList asset parsing."""

from sagemap.assets import MPPositionsList
from .conftest import create_context, load_asset_bytes


def test_mp_positions_list():
    """Test MPPositionsList asset parsing."""
    asset_bytes = load_asset_bytes("MPPositionList")
    
    context = create_context(asset_bytes)
    result = MPPositionsList.parse(context)
    assert result is not None



import logging
from contextlib import contextmanager
from enum import IntEnum
from typing import TypedDict

from .stream import BinaryStream


class AssetPropertyType(IntEnum):
    Boolean = 0
    Integer = 1
    RealNumber = 2
    AsciiString = 3
    UnicodeString = 4
    Unknown = 5


class Property(TypedDict):
    name: str
    type: AssetPropertyType
    value: str | int | float | bool


class ParsingContext:
    def __init__(self, stream: BinaryStream):
        self.stream = stream
        self.assets = {}

        self.compression_bytes = None
        self.asset_count = None

        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.NullHandler())

    def set_logger(self, logger):
        self.logger = logger

    def parse_properties(self):
        properties = []
        property_count = self.stream.readUInt16()
        for _ in range(property_count):
            properties.append(self.parse_asset_property())

        return properties

    def parse_asset_property_key(self):
        property_key_byte = self.stream.readUChar()
        property_key_type = AssetPropertyType(property_key_byte)
        property_key_name_index = self.stream.readUInt24()
        property_key_name = self.assets.get(property_key_name_index)

        return property_key_type, property_key_name_index, property_key_name

    def properties_to_dict(self, properties) -> dict[str, Property]:
        result = {}
        for name, ptype, value in properties:
            if name in result:
                raise ValueError(f"Duplicate property name detected: {name}")
            result[name] = Property(name=name, type=ptype, value=value)
        return result

    def parse_asset_property(self):
        property_key_type, property_key_name_index, property_key_name = self.parse_asset_property_key()
        if property_key_type == AssetPropertyType.Boolean:
            value = self.stream.readBool()
        elif property_key_type == AssetPropertyType.Integer:
            value = self.stream.readInt32()
        elif property_key_type == AssetPropertyType.RealNumber:
            value = self.stream.readFloat()
        elif property_key_type in (AssetPropertyType.AsciiString, AssetPropertyType.Unknown):
            value = self.stream.readUInt16PrefixedAsciiString()
        elif property_key_type == AssetPropertyType.UnicodeString:
            value = self.stream.readUInt16PrefixedUnicodeString()
        else:
            raise ValueError(f"Unexpected property type: {property_key_type}")

        self.logger.debug(
            f"Property: {property_key_name} (Index: {property_key_name_index}), Type: {property_key_type.name}, Value: {value}"
        )
        return (property_key_name, property_key_type, value)

    def parse_assets(self):
        self.compression_bytes = self.stream.readFourCc()
        self.asset_count = self.stream.readUInt32()

        self.assets = {}
        for i in range(self.asset_count, 0, -1):
            asset_name = self.stream.readString()
            asset_index = self.stream.readUInt32()
            if asset_index != i:
                raise ValueError(f"Invalid asset index: expected {i}, got {asset_index}")

            self.assets[asset_index] = asset_name

        return self.assets

    def parse_asset_name(self):
        asset_index = self.stream.readUInt32()
        asset_name = self.assets[asset_index]
        self.logger.debug(f"Parsing asset: {asset_name} (Index: {asset_index})")
        return asset_name

    def parse_asset_header(self):
        asset_version = self.stream.readUInt16()
        datasize = self.stream.readUInt32()
        return asset_version, datasize

    @contextmanager
    def read_asset(self):
        version, datasize = self.parse_asset_header()
        start_pos = self.stream.tell()

        yield version, datasize

        end_pos = self.stream.tell()
        if end_pos - start_pos != datasize:
            raise ValueError(f"Asset data size mismatch: expected {datasize} bytes, read {end_pos - start_pos} bytes")


class WritingContext:
    def __init__(self, stream: BinaryStream):
        self.stream = stream
        self.assets = {}
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.NullHandler())

    @contextmanager
    def write_asset(self, asset_name: str, version: int):
        self.logger.debug(f"Writing asset: {asset_name}, Version: {version}")
        self.stream.writeUInt16(version)
        data_size_position = self.stream.tell()
        self.stream.writeUInt32(0)
        data_position = self.stream.tell()

        yield

        end_position = self.stream.tell()
        data_size = end_position - data_position
        self.stream.seek(data_size_position)
        self.stream.writeUInt32(data_size)
        self.stream.seek(end_position)



"""Test WaypointsList asset parsing."""

from sagemap.assets import WaypointsList
from .conftest import create_context, load_asset_bytes


def test_waypoints_list():
    """Test WaypointsList asset parsing."""
    asset_bytes = load_asset_bytes("WaypointsList")
    
    context = create_context(asset_bytes)
    result = WaypointsList.parse(context)
    assert result is not None



"""Test LibraryMapLists asset parsing."""

from sagemap.assets import LibraryMapLists
from .conftest import create_context, load_asset_bytes


def test_library_map_lists():
    """Test LibraryMapLists asset parsing."""
    asset_bytes = load_asset_bytes("LibraryMapLists")
    
    context = create_context(asset_bytes)
    result = LibraryMapLists.parse(context)
    assert result is not None



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class PostEffectParameter:
    name: str
    type: str
    value: float | tuple[float, float, float, float] | int | str

    @classmethod
    def parse(cls, context: "ParsingContext"):
        param_name = context.stream.readUInt16PrefixedAsciiString()
        param_type = context.stream.readUInt16PrefixedAsciiString()

        if param_type == "Float":
            data = context.stream.readFloat()
        elif param_type == "Float4":
            data = (
                context.stream.readFloat(),
                context.stream.readFloat(),
                context.stream.readFloat(),
                context.stream.readFloat(),
            )
        elif param_type == "Int":
            data = context.stream.readInt32()
        elif param_type == "Texture":
            data = context.stream.readUInt16PrefixedAsciiString()
        else:
            raise ValueError(f"Unknown effect parameter type '{param_type}' for parameter name '{param_name}'.")

        return cls(name=param_name, type=param_type, value=data)


@dataclass
class PostEffect:
    name: str
    parameters: list[PostEffectParameter] | None
    blend_factor: float | None
    lookup_image: str | None

    @classmethod
    def parse(cls, context: "ParsingContext", version: int):
        name = context.stream.readUInt16PrefixedAsciiString()

        parameters = []
        blend_factor = None
        lookup_image = None
        if version >= 2:
            parameter_count = context.stream.readUInt32()
            for _ in range(parameter_count):
                parameters.append(PostEffectParameter.parse(context))
        else:
            blend_factor = context.stream.readFloat()
            lookup_image = context.stream.readUInt16PrefixedAsciiString()

        return cls(
            name=name,
            parameters=parameters if parameters else None,
            blend_factor=blend_factor,
            lookup_image=lookup_image,
        )


@dataclass
class PostEffectsChunk:
    asset_name = "PostEffectsChunk"

    version: int
    post_effects: list[PostEffect]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            post_effects_count = context.stream.readUInt32() if version >= 2 else context.stream.readUChar()
            post_effects = []
            for _ in range(post_effects_count):
                post_effects.append(PostEffect.parse(context, version))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version=version, post_effects=post_effects)



"""Test PlayerScriptsList asset parsing."""

from sagemap.assets import PlayerScriptsList
from .conftest import create_context, load_asset_bytes


def test_player_scripts_list():
    """Test PlayerScriptsList asset parsing."""
    asset_bytes = load_asset_bytes("PlayerScriptsList")
    
    context = create_context(asset_bytes)
    result = PlayerScriptsList.parse(context)
    assert result is not None



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class RiverArea:
    asset_name = "RiverArea"

    unique_id: int
    name: str
    layer_name: str
    uv_scroll_speed: float
    use_additive_blending: bool
    river_texture: str
    noise_texture: str
    alpha_edge_texture: str
    sparkle_texture: str
    color: tuple[int, int, int]
    alpha: float
    water_height: int
    river_type: str | None
    minimum_water_lod: str
    lines: list[tuple[tuple[float, float], tuple[float, float]]]

    @classmethod
    def parse(cls, context: "ParsingContext", version: int):
        unique_id = context.stream.readUInt32()
        name = context.stream.readUInt16PrefixedAsciiString()
        layer_name = context.stream.readUInt16PrefixedAsciiString()
        uv_scroll_speed = context.stream.readSingle()
        use_additive_blending = context.stream.readBool()
        river_texture = context.stream.readUInt16PrefixedAsciiString()
        noise_texture = context.stream.readUInt16PrefixedAsciiString()
        alpha_edge_texture = context.stream.readUInt16PrefixedAsciiString()
        sparkle_texture = context.stream.readUInt16PrefixedAsciiString()
        color = (
            context.stream.readUChar(),
            context.stream.readUChar(),
            context.stream.readUChar(),
        )

        unused_color_a = context.stream.readUChar()
        if unused_color_a != 0:
            raise ValueError(f"Expected unused color alpha to be 0, got {unused_color_a}")

        alpha = context.stream.readSingle()
        water_height = context.stream.readUInt32()

        river_type = None
        if version >= 3:
            river_type = context.stream.readUInt16PrefixedAsciiString()

        minimum_water_lod = context.stream.readUInt16PrefixedAsciiString()

        lines_count = context.stream.readUInt32()
        lines = []

        for _ in range(lines_count):
            lines.append((context.stream.readVector2(), context.stream.readVector2()))

        return cls(
            unique_id=unique_id,
            name=name,
            layer_name=layer_name,
            uv_scroll_speed=uv_scroll_speed,
            use_additive_blending=use_additive_blending,
            river_texture=river_texture,
            noise_texture=noise_texture,
            alpha_edge_texture=alpha_edge_texture,
            sparkle_texture=sparkle_texture,
            color=color,
            alpha=alpha,
            water_height=water_height,
            river_type=river_type,
            minimum_water_lod=minimum_water_lod,
            lines=lines,
        )


@dataclass
class RiverAreas:
    asset_name = "RiverAreas"

    version: int
    areas: list[RiverArea]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            river_area_count = context.stream.readUInt32()
            areas = []
            for _ in range(river_area_count):
                areas.append(RiverArea.parse(context, version))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, areas)



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..context import ParsingContext


@dataclass
class LibraryMaps:
    asset_name = "LibraryMaps"

    version: int
    values: list[str]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, _):
            values_count = context.stream.readUInt32()
            values = []

            for _ in range(values_count):
                values.append(context.stream.readUInt16PrefixedAsciiString())

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, values)


@dataclass
class LibraryMapLists:
    asset_name = "LibraryMapLists"

    version: int
    lists: list[LibraryMaps]

    @classmethod
    def parse(cls, context: "ParsingContext"):
        with context.read_asset() as (version, datasize):
            end_pos = context.stream.tell() + datasize

            lists = []
            while context.stream.tell() < end_pos:
                asset_name = context.parse_asset_name()
                if asset_name != LibraryMaps.asset_name:
                    raise ValueError(f"Expected {LibraryMaps.asset_name} asset, got {asset_name}")

                lists.append(LibraryMaps.parse(context))

        context.logger.debug(f"Finished parsing {cls.asset_name}")
        return cls(version, lists)
